/*
  CPP_CONTEST=2013Cal-2012
  CPP_PROBLEM=R
  CPP_LANG=C+OPENMP+MPI
  CPP_PROCESSES_PER_NODE=mercurio 1 marte 1
*/

/* RECORD
Ángel Conesa Garre
student of Methodology of  Parallel Programming,
at the University of Murcia, Spain
February 1, 2016
time 3397 msec
speed-up 4.80
*/ 

/*
The following code has been improved as follows:
1. The number of threads is set to the number of pairs per process.
2. The first 'for' performed to copy 'cb' in 'copycb' can be avoided by establishing within the 'for' the main calculating the 'copycp' as the first statement.
3. 'MPI_Gatherv' function is more efficient and also avoids all instructions and check 'for' instructions getting more and more readable code being optimized.
*/

#include <stdlib.h>
#include <stdio.h>
#include <mpi.h>
#include <omp.h>


#define MARK 5 //this value in a position in the checkerboard indicates the position has been marked

#ifdef DEBUG
  extern void viewcheckerboard(int,int *);
#endif

int impulsing(int n,int *cb,int ix,int iy,int *fx,int *fy)
{
  int temp;

//if the position is out of the checkerboard do nothing
  if(ix<0 || ix>=n || iy<0 || iy>=n)
  {
    return 0;
  }
  temp=cb[ix*n+iy];
//if the position is marked or an impulsing position do nothing
  if(1<=temp && temp<=MARK)
  {
    return 0;
  }
//if the position is empty, store it and indicate a position has been found
  if(temp==0)
  {
    *fx=ix;
    *fy=iy;
    cb[ix*n+iy]=MARK;
    return 1;
  }
  return 0;
}


int move(int n,int *cb,int ix,int iy,int *fx,int *fy)
{
//if the position is out of the checkerboard do nothing
  if(ix<0 || ix>=n || iy<0 || iy>=n)
  {
    return 0;
  }
//if the position is marked do nothing
  if(cb[ix*n+iy]==MARK)
  {
    return 0;
  }
//if the position is empty, store it and indicate a position has been found
  if(cb[ix*n+iy]==0)
  {
    *fx=ix;
    *fy=iy;
    cb[ix*n+iy]=MARK;
    return 1;
  }
//impulsing to the left
  if(cb[ix*n+iy]==1)
  {
    return(impulsing(n,cb,ix,iy-1,fx,fy));
  }
  else if(cb[ix*n+iy]==2)
  {
    return(impulsing(n,cb,ix-1,iy,fx,fy));
  }
  else if(cb[ix*n+iy]==3)
  {
    return(impulsing(n,cb,ix,iy+1,fx,fy));
  }
  else if(cb[ix*n+iy]==4)
  {
    return(impulsing(n,cb,ix+1,iy,fx,fy));
  }
  return 0;
}

int minimummov(int n,int *cb,int sx,int sy,int tx,int ty)
{
  int *activesx,*activesy; //coordenates of the actives positions
  int nactives; //number of active positions
  int *tempx,*tempy; //auxiliar vector to store temporally the coordenates of the actives positions
  int *temp,i,j;
  int step,end,foundposition;
  char c;

  if(sx==tx && sy==ty)
    return 0;
    
  activesx=(int *) malloc(sizeof(int)*n*n);
  activesy=(int *) malloc(sizeof(int)*n*n);
  tempx=(int *) malloc(sizeof(int)*n*n);
  tempy=(int *) malloc(sizeof(int)*n*n);

  cb[sx*n+sy]=MARK; //the source is marked
  activesx[0]=sx;
  activesy[0]=sy;
  nactives=1;
  step=1;
  end=0;
  do
  {
     i=0;
     j=0;
     while(i<nactives && end==0)
    {
       foundposition=move(n,cb,activesx[i],activesy[i]-1,&tempx[j],&tempy[j]); //moves to the left. If movement is posible the position is stored and the number of stored positions is incremented
#ifdef DEBUG
  printf("Step %d. After moving to the left, from %d,%d, found %d in position %d,%d:\n",step,activesx[i],activesy[i],foundposition,tempx[j],tempy[j]);
  viewcheckerboard(n,cb);
  scanf("%c\n",&c);
#endif
       if(foundposition)
       {
          if(tempx[j]==tx && tempy[j]==ty) //the stored position is the target position
          {
             end=1;
          }
          j++;
       }
      if(end==0)
       {
         foundposition=move(n,cb,activesx[i]-1,activesy[i],&tempx[j],&tempy[j]);
#ifdef DEBUG
  printf("Step %d. After moving up, from %d,%d, found %d in position %d,%d:\n",step,activesx[i],activesy[i],foundposition,tempx[j],tempy[j]);
  viewcheckerboard(n,cb);
  scanf("%c\n",&c);
#endif
         if(foundposition)
         {
            if(tempx[j]==tx && tempy[j]==ty) //the stored position is the target position
            { 
               end=1;
            }
            j++;
         }
       }
       if(end==0)
       {
         foundposition=move(n,cb,activesx[i],activesy[i]+1,&tempx[j],&tempy[j]);
#ifdef DEBUG
  printf("Step %d. After moving to the right, from %d,%d, found %d in position %d,%d:\n",step,activesx[i],activesy[i],foundposition,tempx[j],tempy[j]);
  viewcheckerboard(n,cb);
  scanf("%c\n",&c);
#endif
         if(foundposition)
         {
            if(tempx[j]==tx && tempy[j]==ty) //the stored position is the target position
            {
               end=1;
            }
            j++;
         }
       }
       if (end==0)
       {
         foundposition=move(n,cb,activesx[i]+1,activesy[i],&tempx[j],&tempy[j]);
#ifdef DEBUG
  printf("Step %d. After moving down, from %d,%d, found %d in position %d,%d:\n",step,activesx[i],activesy[i],foundposition,tempx[j],tempy[j]);
  viewcheckerboard(n,cb);
  scanf("%c\n",&c);
#endif
         if(foundposition)
         {
            if(tempx[j]==tx && tempy[j]==ty) //the stored position is the target position
            {
               end=1;
            }
            j++;
         }
       }
      i++;
    } //while
    nactives=j;
    temp=activesx;
    activesx=tempx;
    tempx=temp;
    temp=activesy;
    activesy=tempy;
    tempy=temp;
    step++;
  } while(nactives && !end);
  step--;
  free(activesx);
  free(activesy);
  free(tempx);
  free(tempy);
#ifdef DEBUG
  printf("Minimo de %d,%d a %d,%d: %d\n\n",sx,sy,tx,ty,step);
#endif
  return step;
}

void sec(int n,int *cb,int p,int *positions,int *movements,int node,int np)
{
  if (node != 0) {
    cb = (int *) malloc (sizeof(int)*n*n);
    positions = (int *) malloc (sizeof(int)*4*p);
  }

  MPI_Bcast(cb, n*n, MPI_INT, 0, MPI_COMM_WORLD);			// Positions and matrix are distributed to all nodes
  MPI_Bcast(positions, 4*p, MPI_INT, 0, MPI_COMM_WORLD);

  int i,j; //auxiliar

//for each pair of points the minimum number of movements to go from the source to the target is obtained
  int *parejasNodo = (int *) malloc(np*sizeof(int));
  int *desplazamiento = (int *) malloc(np*sizeof(int));
  
  desplazamiento[0]=0;										//  Couples and displacement calculation
  for (i=0; i<np; i++) {
    parejasNodo[i] = (p/np) + ((i<(p%np))?1:0);
    if (i > 0)
      desplazamiento[i] = desplazamiento[i-1] + parejasNodo[i-1];
  }

  int *resultado = (int *) malloc(sizeof(int)*parejasNodo[node]);	// Array for the results of each node
  
  int nThreads=p/np;						// Many threads as couples
  omp_set_num_threads(nThreads);
  #pragma omp parallel private(i,j)
  {
    int *copycb=(int *) malloc(sizeof(int)*n*n);

    #pragma omp for schedule(dynamic)
    for(i=0; i<parejasNodo[node]; i++)    // Calculation
    {
      for(j=0;j<n*n;j++) copycb[j]=cb[j];
      resultado[i]=minimummov(n,copycb,positions[(desplazamiento[node]+i)*4],positions[(desplazamiento[node]+i)*4+1],positions[(desplazamiento[node]+i)*4+2],positions[(desplazamiento[node]+i)*4+3]);
    }
    free(copycb);
  }
  
  MPI_Gatherv(resultado, parejasNodo[node], MPI_INT, movements, parejasNodo, desplazamiento, MPI_INT, 0, MPI_COMM_WORLD); // Combining results

  if (node != 0) 
  {
    free(positions);
    free(cb);
  }
  
  free(resultado);
  free(parejasNodo);
  free(desplazamiento);
}
